@@ -11,7 +11,7 @@
 
 import logging
 
-from typing import Any, Dict, List, Optional, Set, Tuple, Union
+from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union
 
 import os
 import json
@@ -397,7 +397,7 @@ class XEP_0384(BasePlugin):
                 # TODO: Slixmpp should handle pubsub#errors so we don't have to
                 # fish the element ourselves
                 precondition = exn.iq['error'].xml.find(
-                    '{%s}%s' % (PUBSUB_ERRORS, 'precondition-not-met'),
+                    f'{{{PUBSUB_ERRORS}}}precondition-not-met'
                 )
                 if precondition is not None:
                     log.debug('The node we tried to publish was already '
@@ -415,19 +415,30 @@ class XEP_0384(BasePlugin):
             else:
                 log.debug('Not publishing.')
 
-    async def _fetch_bundle(self, jid: str, device_id: int) -> Optional[ExtendedPublicBundle]:
+    async def _fetch_bundle(self, jid: JID, device_id: int) -> None:
+        """
+            Fetch bundles for specified jid / device_id pair.
+        """
         log.debug('Fetching bundle for JID: %r, device: %r', jid, device_id)
-        node = '%s:%d' % (OMEMO_BUNDLES_NS, device_id)
+        node = f'{OMEMO_BUNDLES_NS}:{device_id}'
         try:
             iq = await self.xmpp['xep_0060'].get_items(jid, node)
         except (IqError, IqTimeout):
             return None
         bundle = iq['pubsub']['items']['item']['bundle']
 
-        return _parse_bundle(self.omemo_backend, bundle)
+        bundle = _parse_bundle(self.omemo_backend, bundle)
+        if bundle is not None:
+            log.debug('Encryption: Bundle %r found!', device_id)
+            devices = self.bundles.setdefault(jid.bare, {})
+            devices[device_id] = bundle
+        else:
+            log.debug('Encryption: Bundle %r not found!', device_id)
 
-    async def _fetch_device_list(self, jid: JID) -> None:
-        """Manually query PEP OMEMO_DEVICES_NS nodes"""
+    async def _fetch_devices(self, jid: JID) -> None:
+        """
+            Manually query PEP OMEMO_DEVICES_NS nodes
+        """
         log.debug('Fetching device list for JID: %r', jid)
         iq = await self.xmpp['xep_0060'].get_items(jid.full, OMEMO_DEVICES_NS)
         return await self._read_device_list(jid, iq['pubsub']['items'])
@@ -478,7 +489,7 @@ class XEP_0384(BasePlugin):
                 return  # XXX: Handle this!
 
         if device_ids is None:
-            device_ids = await self.get_device_list(own_jid)
+            device_ids = await self.get_active_devices(own_jid)
 
         devices = []
         for i in device_ids:
@@ -527,10 +538,19 @@ class XEP_0384(BasePlugin):
                     own_jid.bare, OMEMO_DEVICES_NS, payload=payload,
                 )
 
-    async def get_device_list(self, jid: JID) -> List[str]:
-        """Return active device ids. Always contains our own device id."""
+    async def get_devices(self, jid: JID) -> Iterable[str]:
+        """
+            Get all devices for a JID.
+        """
         devices = await self._omemo().getDevices(jid.bare)
-        return devices.get('active', [])
+        return set(devices.get('active', []) + devices.get('inactive', []))
+
+    async def get_active_devices(self, jid: JID) -> Iterable[str]:
+        """
+            Return active device ids. Always contains our own device id.
+        """
+        devices = await self._omemo().getDevices(jid.bare)
+        return set(devices.get('active', []))
 
     async def _should_heartbeat(self, jid: JID, device_id: int, prekey: bool) -> bool:
         """
@@ -591,7 +611,7 @@ class XEP_0384(BasePlugin):
 
         return await self._should_heartbeat(jid, sid, prekey)
 
-    async def make_heartbeat(self, jid: JID, device_id: int) -> Message:
+    async def send_heartbeat(self, jid: JID, device_id: int) -> None:
         """
             Returns a heartbeat message.
 
@@ -610,7 +630,13 @@ class XEP_0384(BasePlugin):
         )
         msg.append(encrypted)
         msg.enable('store')
-        return msg
+        msg.send()
+
+    async def delete_session(self, jid: JID, device_id: int) -> None:
+        """
+            Delete the session for the provided jid/device_id pair.
+        """
+        await self._omemo().deleteSession(jid.bare, device_id)
 
     async def trust(self, jid: JID, device_id: int, ik: bytes) -> None:
         await self._omemo().setTrust(jid.bare, device_id, ik, True)
@@ -641,8 +667,8 @@ class XEP_0384(BasePlugin):
 
         return await self._omemo().getTrustForJID(jid.bare)
 
-    @classmethod
-    def is_encrypted(cls, msg: Message) -> bool:
+    @staticmethod
+    def is_encrypted(msg: Message) -> bool:
         return msg.xml.find('{%s}encrypted' % OMEMO_BASE_NS) is not None
 
     async def decrypt_message(
@@ -720,11 +746,7 @@ class XEP_0384(BasePlugin):
             should_heartbeat = await self._should_heartbeat(sender, sid, isPrekeyMessage)
             if should_heartbeat:
                 log.debug('Decryption: Sending hearbeat to %s / %d', jid, sid)
-                async def send_heartbeat():
-                    log.debug('Sending a heartbeat message')
-                    msg = await self.make_heartbeat(JID(jid), sid)
-                    msg.send()
-                asyncio.ensure_future(send_heartbeat())
+                await self.send_heartbeat(JID(jid), sid)
 
         return body
 
@@ -782,10 +804,11 @@ class XEP_0384(BasePlugin):
                 elif _ignore_trust:
                     if not _device_id or len(barejids) != 1:
                         raise ErroneousParameter
+                    bundle = self.bundles.get(barejids[0], {}).get(_device_id, None)
                     encrypted = await self._omemo().encryptRatchetForwardingMessage(
                         bare_jid=barejids[0],
                         device_id=_device_id,
-                        bundle=None,
+                        bundle=bundle,
                     )
                 else:
                     encrypted = await self._omemo().encryptKeyTransportMessage(
@@ -806,16 +829,10 @@ class XEP_0384(BasePlugin):
             for exn in errors:
                 if isinstance(exn, omemo.exceptions.NoDevicesException):
                     log.debug('Encryption: Missing device list for JID: %r', exn.bare_jid)
-                    await self._fetch_device_list(JID(exn.bare_jid))
+                    await self._fetch_devices(JID(exn.bare_jid))
                 elif isinstance(exn, omemo.exceptions.MissingBundleException):
                     log.debug('Encryption: Missing bundle for JID: %r, device: %r', exn.bare_jid, exn.device)
-                    bundle = await self._fetch_bundle(exn.bare_jid, exn.device)
-                    if bundle is not None:
-                        log.debug('Encryption: Bundle %r found!', exn.device)
-                        devices = self.bundles.setdefault(exn.bare_jid, {})
-                        devices[exn.device] = bundle
-                    else:
-                        log.debug('Encryption: Bundle %r not found.', exn.device)
+                    await self._fetch_bundle(JID(exn.bare_jid), exn.device)
                 elif isinstance(exn, omemo.exceptions.TrustException):
                     # On TrustException, there are two possibilities.
                     # Either trust has not been explicitely set yet, and is
